%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Blocking
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(blocking,
	[blocking_clause_set/2
     ]).

:- use_module('../yarralumla/misc').
:- use_module('../yarralumla/data_structures').


blocking_clause_set(CS, Res) :-
	scan_function_symbols(CS, FSig),
	subterm_clauses(FSig, SubtermClauses),
 	scan_predicate_symbols(CS, PSig),
%% 	speculate_equality_clauses(PSig, EQClauses),
%% No! It is enough to 'check loops' on the domain-predicates alone:
%% If there is no loop through the domain predicates, there is no loop
%% at all. 
%% Yes, but the extension of the domain predicate is pretty large, usually
%% Check for all other predicates:
        delete(PSig, dom/1, PSig1),
	%% and we don't do it for equality:
        delete(PSig1, (=)/2, PSig2),
 	speculate_equality_clauses(PSig2, EQClauses),
	append([CS,SubtermClauses,EQClauses], Res).

subterm_clauses(Sig, [([subterm(X,X)] :- [dom(X)])|Res]) :-
	%% Sig is a list of function symbols plus arities (F/N)
	maplist(subterm_clauses_, Sig, HRes),
	%% Have a list of lists now, flatten it
	append(HRes, Res).

subterm_clauses_(F/N, Res) :-
	%% See example below to make sense of this
	make_N_variables(N, NVars),
	FNVars =.. [F|NVars],
	findall(([subterm(X,FNVars)] :- 
                        [subterm(X,Y), dom(X), dom(FNVars)]),
	        member(Y, NVars),
		Res).

speculate_equality_clauses(PSig, [([] :- [nonequal(X,X)]), 
	                          ([] :- [(Y=X), nonequal(Y,X)])|
			         Res]) :-
	%% Sig is a list of function symbols plus arities (F/N)
	%% PSig is a list of predicate symbols plus arities (P/M)
	%% we check loops wrt all argument positions 1,...,M for
	%% atoms with predicate symbol P
	findall(([Y=X, nonequal(Y,X)] :- [PNXVars, PNYVars, subterm(X,Y)]),
	        ( member(P/N, PSig),
		  make_N_variables(N, NXVars),
		  make_N_variables(N, NYVars),
		  PNXVars =.. [P|NXVars],
		  PNYVars =.. [P|NYVars],
		  member2(X, Y, NXVars, NYVars)
	         ),
		 Res).
	

/*
 * Example:

subterm(X,X) :- dom(X).

subterm(X,f(Y)) :- subterm(X,Y), dom(X), dom(f(Y)).

subterm(X,g(Y1,Y2)) :- subterm(X,Y1), dom(X), dom(g(Y1,Y2)).
subterm(X,g(Y1,Y2)) :- subterm(X,Y2), dom(X), dom(g(Y1,Y2)).

false :- Y=X, nonequal(Y,X).

Y1=X1 ; nonequal(Y1,X1) :- r(X1,X2), r(Y1,Y2), subterm(X1,Y1).
Y2=X2 ; nonequal(Y2,X2) :- r(X1,X2), r(Y1,Y2), subterm(X2,Y2).


 */
