:- module(call_graph,
	[cyclic_definitions/3
     ]).

:- use_module('../yarralumla/misc').
:- use_module('../yarralumla/data_structures').


call_graph_blacklist([(=)/2, my_equal/2, nonequal/2, dom/1, '$krh_dom'/1]).

cyclic_definitions(CS, CyclicUnary, CyclicNAry) :-
        msg('Computing call graph...',[]),
        call_graph_(CS, Calls),
        vertices_edges_to_ugraph([], Calls, Graph),
        transitive_closure(Graph, CGraph),
        vertices(Graph, Preds),
        % return those who are cyclic
        findall(P/1,
                ( member(P/1 - R, CGraph),
                  memberchk(P/1, R)),
                CyclicUnary 
               ),
        findall(P/N,
                ( member(P/N - R, CGraph),
                  N > 1,
                  memberchk(P/N, R)),
                CyclicNAry
               ),
        length(CyclicUnary, NCU),
        length(CyclicNAry, NCN),
        length(Preds, NP),
        msg('Number of predicate symbols: ~w.', [NP]),
        msg('   Thereof unary with cyclic definitions: ~w.', [NCU]),
        msg('   Thereof n-ary (n>1) with cyclic definitions: ~w.', [NCN]).

call_graph_([],[]).
call_graph_([(H :- B)|R],Res) :-
        call_graph_blacklist(BL),
        findall((KPred/KArity-LPred/LArity),
                ( member(K, H), 
                  member(L, B),
                  functor(K, KPred, KArity),
                  functor(L, LPred, LArity),
                   \+ (memberchk(KPred/KArity, BL)),
                   \+ (memberchk(LPred/LArity, BL))
                ), 
                Calls),
        call_graph_(R,RRes),
        append(Calls, RRes, Res).
