%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Flattening
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(flattening_for_shifting, [flatten_clause_set_for_shifting/2]).

:- use_module('../yarralumla/misc').
:- use_module('../yarralumla/data_structures').

flatten_clause_set_for_shifting(CS, [([XeqX] :- [])|CSA]) :-
	%% Notice we add reflexivity.
	%% This gives us a complete transformation for the case
	%% that the input does not contains equality.
	make_unify(X,X,XeqX),
	maplist(flatten_clause_for_shifting, CS, CSAL),
	%% Have a list of list of clauses now, flatten it:
	append(CSAL, CSA).


flatten_clause_for_shifting((H :- B), [H :- BRes]) :-
	flatten_list_for_shifting(B, BFlat, BDefs),
	append([BFlat, BDefs], BRes).

%% lifting of flatten/3 to lists:
flatten_list_for_shifting([],[],[]).
flatten_list_for_shifting([A|R], [AFlat|RFlat], ARDefs) :-
	flatten_for_shifting(A, AFlat, ADefs),
	flatten_list_for_shifting(R, RFlat, RDefs),
	append(ADefs, RDefs, ARDefs).

%% Flatten a single atom:
flatten_for_shifting(A, AFlat, ADefs) :-
	A =.. [P|Args],
 	flatten_term_list_for_shifting(Args, ArgsFlat, ADefs),
	AFlat =.. [P|ArgsFlat].

flatten_term_list_for_shifting([], [], []).
flatten_term_list_for_shifting([T|R], [T|RFlat], RDefs) :-
	( var(T) ; atomic(T)), !,
	flatten_term_list_for_shifting(R, RFlat, RDefs).
flatten_term_list_for_shifting([T|R], [X|RFlat], [XeqT|RDefs]) :-
	make_unify(X,T,XeqT),
	flatten_term_list_for_shifting(R, RFlat, RDefs).



