%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Sort heads
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(sort_heads,
	[sort_heads/2
     ]).

:- use_module('../yarralumla/misc').
:- use_module('../yarralumla/data_structures').
:- use_module('../yarralumla/call_graph').

cnt_blacklist([(=)/2, my_equal/2, nonequal/2, dom/1, '$krh_dom'/1]).


sort_heads(CS, Res) :-
        scan_predicate_symbols(CS, PSig),
%        cnt_blacklist(CBL),
        findall(P/A-[0,0,0],
                member(P/A, PSig),
                ALInit),
        list_to_assoc(ALInit, AInit),
        clauses_pred_occurences(CS, AInit, AOccurences),
        %writeln(sort_clauses_heads(CS, AOccurences, Res)),
        sort_clauses_heads(CS, AOccurences, Res).

sort_clauses_heads([], _, []).
sort_clauses_heads([F|R], PWeights, [FRes|RRes]) :-
        sort_clause_head(F, PWeights, FRes),
        %writeln(sort_clause_head(F, PWeights, FRes)),
        sort_clauses_heads(R, PWeights, RRes).

sort_clause_head(([] :- B), _PWeights, ([] :- B)) :- !.
sort_clause_head(([A] :- B), _PWeights, ([A] :- B)) :- !.
sort_clause_head(([(X=Y), nonequal(X,Y)] :- B), _PWeights, 
                 ([(X=Y), nonequal(X,Y)] :- B)) :- !.
sort_clause_head((H :- B), PWeights, (HS :- B)) :-
        attach_weights(H, PWeights, HW),
        keysort(HW, HWS),
        detach_weights(HWS, HS).

attach_weights([],_,[]).
attach_weights([A|R], PWeights, [W-A|RRes]) :-
        functor(A, Pred, Arity),
        get_assoc(Pred/Arity, PWeights, W),
        attach_weights(R, PWeights, RRes).

detach_weights([],[]).
detach_weights([_W-A|R], [A|RRes]) :-
        detach_weights(R, RRes).


clauses_pred_occurences([], A, A).
clauses_pred_occurences([F|R], AIn, AOut) :-
        clause_head_pred_occurences(F, AIn, HA1),
        clause_body_pred_occurences(F, HA1, HA2),
        clauses_pred_occurences(R, HA2, AOut).


clause_head_pred_occurences((H :- _B), AIn, AOut) :-
        clause_head_pred_occurences_(H, AIn, AOut).

clause_head_pred_occurences_([], A, A).
clause_head_pred_occurences_([F|R], AIn, AOut) :-
        functor(F, Pred, Arity),
        cnt_blacklist(CBL),
        memberchk(Pred/Arity, CBL), !,
        clause_head_pred_occurences_(R, AIn, AOut).
clause_head_pred_occurences_([F|R], AIn, AOut) :-
        functor(F, Pred, Arity),
        get_assoc(Pred/Arity, AIn, [_, PBodyCnt, PHeadCnt]),
        NewPHeadCnt is PHeadCnt+1,
        Sum is NewPHeadCnt + PBodyCnt,
        put_assoc(Pred/Arity, AIn, [Sum, PBodyCnt, NewPHeadCnt], HA),
        clause_head_pred_occurences_(R, HA, AOut).

clause_body_pred_occurences((_H :- B), AIn, AOut) :-
        clause_body_pred_occurences_(B, AIn, AOut).

clause_body_pred_occurences_([], A, A).
clause_body_pred_occurences_([F|R], AIn, AOut) :-
        functor(F, Pred, Arity),
        cnt_blacklist(CBL),
        memberchk(Pred/Arity, CBL), !,
        clause_body_pred_occurences_(R, AIn, AOut).
clause_body_pred_occurences_([F|R], AIn, AOut) :-
        functor(F, Pred, Arity),
        get_assoc(Pred/Arity, AIn, [_, PBodyCnt, PHeadCnt]),
        NewPBodyCnt is PBodyCnt+1,
        Sum is NewPBodyCnt + PHeadCnt,
        put_assoc(Pred/Arity, AIn, [Sum, NewPBodyCnt, PHeadCnt], HA),
        clause_body_pred_occurences_(R, HA, AOut).

        
        

        