%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Standard Range restriction
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(standard_range_restriction,
	[standard_rr_clause_set/2,
	standard_rr_clause/2
     ]).

:- use_module('../yarralumla/misc').
:- use_module('../yarralumla/data_structures').
%% :- use_module('../yarralumla/sorting').


standard_rr_clause_set(CSH, Res) :-
	%% Add reflexivity - it is the only equality axiom that
	%% will be modified by the transformation
	CS = [([X=X]:-[dom(X)])|CSH],
	%% (1) add a constant if none is there
 	scan_function_symbols(CS, FSig),
	( member(AConstant/0, FSig) ; AConstant = a_constant ),
	Res1 = [([dom(AConstant)] :- [])],
	%% (2) Make the standard RR axioms:
	standard_rr_axioms(FSig, Res2),
	%% (3) Range-restrict the given clause set:
	maplist(standard_rr_clause, CS, Res3),
	%% Finally collect the grand result:
	append([Res1, Res2, Res3], Res).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

standard_rr_clause((H :- B), (H :- BDom)) :-
	term_variables(H, HVars),
	term_variables(B, BVars),
	%% For *extra* variables  ...
	subtractq(HVars, BVars, HDomVars), 
	make_doms(HDomVars, DomVarsDom),
	append(B, DomVarsDom, BDom).

make_doms([],[]).
make_doms([X|R],[dom(X)|DomR]) :- make_doms(R,DomR).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

standard_rr_axioms(Sig, Res) :-
	%% Sig is a list of function symbols plus arities (F/N)
	maplist(standard_rr_axioms_1, Sig, Res).

standard_rr_axioms_1(F/N, ([dom(FNXVars)] :- NXVarsDom)) :-
	%% See example below to make sense of this
	make_N_variables(N, NXVars),
	FNXVars =.. [F|NXVars],
	make_doms(NXVars, NXVarsDom).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

dom_subterm_clauses(Sig, Res) :-
	%% Sig is a list of function symbols plus arities (F/N)
	maplist(dom_subterm_clauses_1, Sig, HRes),
	%% Have a list of lists now, flatten it
	append(HRes, Res).

dom_subterm_clauses_1(F/N, Res) :-
	%% See example below to make sense of this
	make_N_variables(N, NXVars),
	FNXVars =.. [F|NXVars],
	findall(([dom(X)] :- [dom(FNXVars)]),
	        member(X, NXVars),
		Res).
